#ifndef __LMP_H
#define __LMP_H

// #define DEBUG

#ifndef __C2MAN__
#include <stdlib.h>
#endif

#define LMP_VERSION VERSION

#define LMP_ARG_MAX_LENGTH   100
#define LMP_MISSING_ARG      1

#define LMP_NO_ERROR         0
#define LMP_UNKNOWN_OPERATOR 1

/* lmp uses as standard type for numbers */
typedef long double lmp_value_type;
/* operators lmp knows */
typedef enum lmp_op_enum {
  LMP_OUTER_OR,
  LMP_OUTER_AND,
  LMP_EQUALS,
  LMP_UNEQUALS,
  LMP_GREATER_EQUAL,
  LMP_LOWER_EQUAL,
  LMP_GREATER,
  LMP_LOWER,
  LMP_ADDITION,
  LMP_SUBSTRACTION,
  LMP_MULTIPLICATION,
  LMP_DIVISION,
  LMP_MODULO,
  LMP_DIV_NO_REM,
  LMP_POWER,
  LMP_LOG,
  LMP_SINE,
  LMP_COSINE,
  LMP_TANGENT,
  LMP_COTANGENT,
  LMP_ARC_SINE,
  LMP_ARC_COSINE,
  LMP_ARC_TANGENT,
  LMP_LOGARITHM,
  LMP_LOG_10,
  LMP_EXP,
  LMP_SQRT,
} lmp_operator;
/* var table entry */
typedef struct {
  char *         var;
  lmp_value_type val;
} var_table_t;

extern char ** lmp_args;
extern int lmp_argnum;
extern var_table_t * lmp_vt;
extern int lmp_vt_size;
extern unsigned short lmp_proc_error;

#ifndef __C2MAN__
extern long double strtold __P ((__const char * __nptr, char **__endptr));
#endif

void lmp_no_args ();
void lmp_print_help ();
void lmp_print_version ();

char * lmp_ver ();

void lmp_stop (const char * msg);
void lmp_error (const char * msg, int arg);

char * lmp_next_token ();
char * lmp_following_token ();
char * lmp_last_token ();
int lmp_inc_token ();
int lmp_get_arg_num ();
int lmp_following_equals (char *s);
lmp_value_type lmp_next_number ();

int lookup_var_table (const char * s, lmp_value_type * val);
char ** lmp_parse_string_by_space (const char * string, int * argc);
char ** lmp_parse_string_by_elements (char * string, int * argc);
int init_var_table (const char * vs);

lmp_value_type lmp_eval ();

lmp_value_type lmp_binary_proc (lmp_operator op, lmp_value_type left, lmp_value_type right);
lmp_value_type lmp_unary_proc (lmp_operator op, lmp_value_type operand);

char * base2base (unsigned base, long value, int precision);
char * dyn_format (const char * start, int value1, int value2, const char * end);

#endif
