#include "lmp.h"
#ifdef DEBUG
#include <stdio.h>
#endif
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include "lmpinput.h"

/* vector of arguments for lmp */
char ** lmp_args = NULL;
/* index for lmp_args */
int lmp_argnum = 1;

/* var table */
var_table_t * lmp_vt = NULL;
/* var table size */
int lmp_vt_size = 0;

/* scans the input for the next token
 * returns next token
 */
char * lmp_next_token () {
  static char arg [LMP_ARG_MAX_LENGTH];

  if (*lmp_args == NULL)
#ifdef DEBUG
  {
    fprintf (stderr, "(DBG) next_token: No more tokens.\n");
#endif
    return NULL;
#ifdef DEBUG
  }
#endif

  strcpy (arg, *lmp_args++);
  lmp_argnum++;
#ifdef DEBUG
  fprintf (stderr, "(DBG) next_token: %s.\n", arg);
#endif
  return arg;
}

/* Increments the token pointer.
 * It does not return a token.
 */
int lmp_inc_token () {
  if (*lmp_args++ == NULL)
#ifdef DEBUG
  {
    fprintf (stderr, "(DBG) inc_token: No more arguments.\n");
#endif
    return LMP_MISSING_ARG;
#ifdef DEBUG
  }
#endif
  else
#ifdef DEBUG
  {
    fprintf (stderr, "(DBG) inc_token: increase argnum\n");
#endif
    lmp_argnum++;
#ifdef DEBUG
  }
#endif

  return 0;
}

/* Get the number of the current argument
 */
int lmp_get_arg_num () {
#ifdef DEBUG
  fprintf (stderr, "(DBG) get_arg_num: %d\n", lmp_argnum);
#endif
  return lmp_argnum;
}

/* Compares a string with the next token
 * It compares the current token with "s", but it does not
 * increment the token pointer.
 * returns 1 if equal and 0 if not.
 */
int lmp_following_equals (s)
char * s; /* to be compared string */
{
  if (*lmp_args == NULL)
#ifdef DEBUG
  {
    fprintf (stderr, "(DBG) following_equals: No more arguments to compare with %s.\n", s);
#endif
    return 0;
#ifdef DEBUG
  }
#endif
  else if (!strcmp (s, lmp_following_token ()))
#ifdef DEBUG
  {
    fprintf (stderr, "(DBG) following_equals: argument matches \"%s\".\n", s);
#endif
    return 1;
#ifdef DEBUG
  }
#endif
  else
#ifdef DEBUG
  {
    fprintf (stderr, "(DBG) following_equals: argument does not matches \"%s\".\n", s);
#endif
    return 0;
#ifdef DEBUG
  }
#endif
}

/* watch the following token
 * Does not increment the token pointer but returns
 * the following token.
 */
char * lmp_following_token () {
  static char arg [LMP_ARG_MAX_LENGTH];

  if (*lmp_args == NULL)
#ifdef DEBUG
  {
    fprintf (stderr, "(DBG) following_token: No more arguments.\n");
#endif
    return NULL;
#ifdef DEBUG
  }
#endif

  strcpy (arg, *lmp_args);
#ifdef DEBUG
  fprintf (stderr, "(DBG) following_token: Got \"%s\".\n", arg);
#endif
  return arg;
}

/* Returns the last token.
 */
char * lmp_last_token () {
  static char arg [LMP_ARG_MAX_LENGTH];

  if (*(lmp_args-1) == NULL)
#ifdef DEBUG
  {
    fprintf (stderr, "(DBG) last_token: No last argument.\n");
#endif
    return NULL;
#ifdef DEBUG
  }
#endif

  strcpy (arg, *(lmp_args-1));
#ifdef DEBUG
  fprintf (stderr, "(DBG) following_token: Got \"%s\".\n", arg);
#endif
  return arg;
}

/* Return a number
 * Reads a token. It expects a number. If the token is something else,
 * lmp will stop.
 * returns the value of the number.
 */
lmp_value_type lmp_next_number () {
  char *s = lmp_next_token (), *endptr;
  lmp_value_type re;

  if (!s)
#ifdef DEBUG
  {
    fprintf (stderr, "(DBG) next_number: Number missed.\n");
#endif
    lmp_error ("At argument %d value missed.\n", 0);
#ifdef DEBUG
  }
#endif

  if (lookup_var_table (s, &re)) {
    re = STRTOFVAL (s, &endptr);

    if (*endptr != 0 && re == 0)
#ifdef DEBUG
      {
	fprintf (stderr, "(DBG) next_number: Number expected but not found\n");
#endif
	lmp_error ("At argument %d value expected.\n", -1);
#ifdef DEBUG
      }
#endif
  }

#ifdef DEBUG
  fprintf (stderr, "(DBG) next_number: token \"%s\" result is %Lf last char \"%c\"\n", s, re, *endptr);
#endif
  return re;
}

/* Parse string to an array of strings.
 * lmp_parse_string() parses string. argc is set to the
 * number of elements lmp_parse_string() will found.
 *
 * This function may be used before lmp_eval().
 * returns the strings vector.
 */
char ** lmp_parse_string_by_space (string,argc)
const char * string; /* string to be parsed */
int * argc;          /* returns the number of elements */
{
  char ** argv, * sp, * tok;

  argv = NULL;
  *argc = 0;

  sp = malloc (strlen (string) + 1);
  strcpy (sp, string);

  while ((tok = strsep (&sp, " ")) != NULL) {
    (*argc)++;
    argv = realloc (argv, *argc * sizeof (char **));
    argv [*argc-1] = malloc (strlen (tok) + 1);
    strcpy (argv [*argc-1], tok);
  }
  free (sp);

  return argv;
}

/* Parse string to an array by scanning
 * returns the string vector
 */
char ** lmp_parse_string_by_elements (s,c) 
char * s;
int * c;
{
  enum { NOTHING, DIGIT, SPACE, SYMBOL, ALPHA, PLUS, MINUS, ASTERISK, AND, OR, EQUAL, LOWER, GREATER, NOT } state;
  char ** res, * last;
  int len, i, str;

  str = len = 0;
  state = NOTHING;
  res = malloc ((str+1)*sizeof (char *));
  last = s;
  for (i = 0; i < strlen (s); i++) {
    if ((s[i] >= '0' && s[i] <= '9') || (s[i] == '.')) { // Digit
      if (state != DIGIT) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
	state = DIGIT;
      }
      len++;
    }
    else if ((s[i] == ' ') || (s[i] == '\t') || (s[i] == '\n')) { // Space
      if (state != NOTHING) {
	res[str] = malloc (len + 1);
	strncpy (res[str], last, len);
	res[str++][len] = '\0';
	res = realloc (res, (str+1)*sizeof (char *));
      }
      last = s+i+1;
      state = NOTHING;
      len = 0;
    }
    else if ((s[i] == '(') || (s[i] == ')') || (s[i] == '/') || (s[i] == '~')) { // Symbol
      if (state != NOTHING) {
	res[str] = malloc (len + 1);
	strncpy (res[str], last, len);
	res[str++][len] = '\0';
	res = realloc (res, (str+1)*sizeof (char *));
	last = s+i;
      }
      len = 1;
      state = SYMBOL;
    }
    else switch (s[i]) {
    case '*':
      if (state != ASTERISK) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
	state = ASTERISK;
      }
      len++;
      break;
    case '+':
      if (state != PLUS) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
	state = PLUS;
      }
      len++;
      break;      
    case '|':
      if (state != OR) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
	state = OR;
      }
      len++;
      break;
    case '&':
      if (state != AND) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
	state = AND;
      }
      len++;
      break;      
    case '-':
      if (state != MINUS) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
	state = MINUS;
      }
      len++;
      break;      
    case '=':
      if ((state != EQUAL) && (state != PLUS) && (state != MINUS) && (state != ASTERISK) && (state != AND) && (state != OR) && (state != LOWER) && (state != GREATER) && (state != NOT)) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
	state = EQUAL;
      }
      len++;
      break;      
    case '!':
      if (state != NOT) {
	if (state != NOTHING) {
	  res[str] = malloc (len + 1);
	  strncpy (res[str], last, len);
	  res[str++][len] = '\0';
	  res = realloc (res, (str+1)*sizeof (char *));
	  last = s+i;
	}
	len = 0;
	state = NOT;
      }
      len++;
      break;      
    default: { // Alphas
	if (state != ALPHA) {
	  if (state != NOTHING) {
	    res[str] = malloc (len + 1);
	    strncpy (res[str], last, len);
	    res[str++][len] = '\0';
	    res = realloc (res, (str+1)*sizeof (char *));
	    last = s+i;
	  }
	  len = 0;
	  state = ALPHA;
	}
	len++;
      }
      break;
    }
  }

  if (state != NOTHING) {
    res[str] = malloc (len + 1);
    strncpy (res[str], last, len);
    res[str++][len] = '\0';
    last = s+i;
    len = 0;
  }

  res [str] = NULL;
  *c = str;

  return res;
}

/* search a variable in the var table
 * returns 0 if successful, else 1
 */
int lookup_var_table (s, val)
const char * s;       /* to be compared string */
lmp_value_type * val; /* return value */
{
  int i;

  for (i = 0; i < lmp_vt_size; i++)
    if (!strcmp (s, lmp_vt [i].var)) {
      *val = lmp_vt [i].val;
      return 0;
    }

  return 1;
}

/* initialize the var table
 * returns 0 if successful
 */
int init_var_table (vs)
const char * vs; /* initializer string: "x=3:y=-0.5..." */
{
  char * sp, * tok, * var, * val, * endptr;

  if (lmp_vt)
    free (lmp_vt);

  lmp_vt_size = 0;
  
#ifdef DEBUG
  fprintf (stderr, "(DBG) init_var_table: %s.\n", vs);
#endif

  sp = malloc (strlen (vs) + 1);
  strcpy (sp, vs);

  while ((tok = strsep (&sp, ":")) != NULL) {
    lmp_vt_size++;
    lmp_vt = realloc (lmp_vt, lmp_vt_size * sizeof (var_table_t));
    var = strsep (&tok, "=");
    val = strsep (&tok, "=");
    lmp_vt [lmp_vt_size-1].var = malloc (strlen (var) + 1);
    strcpy (lmp_vt [lmp_vt_size-1].var, var);
    lmp_vt [lmp_vt_size-1].val = STRTOFVAL (val, &endptr);

    if (*endptr != 0 && lmp_vt [lmp_vt_size-1].val == 0)
#ifdef DEBUG
      {
	fprintf (stderr, "(DBG) init_var_table: Number expected but not found\n");
#endif
	lmp_error ("At argument %d value expected.\n", -1);
#ifdef DEBUG
      }
#endif
  }

  if (*endptr != 0 && lmp_vt [lmp_vt_size-1].val == 0)
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) init_var_table: Number expected but not found\n");
#endif
      lmp_error ("At argument %d value expected.\n", -1);
#ifdef DEBUG
    }
#endif
  free (sp);

  return 0;
}
